@@ -17,20 +17,12 @@
 --  This Module contains test cases for module URI.
 --
 --  To run this test without using Cabal to build the package
---  (2013-01-05, instructions tested on MacOS):
+--  (2013-01-05, instructions tested on macOS):
 --  1. Install Haskell platform
---  2. cabal install test-framework
---  3. cabal install test-framework-hunit
---  4. ghc -XDeriveDataTypeable -D"MIN_VERSION_base(x,y,z)=1" ../Network/URI.hs uri001.hs
+--  2. cabal install tasty tasty-hunit tasty-quickcheck QuickCheck
+--  3. ghc -XDeriveDataTypeable -XDeriveGeneric -package QuickCheck -package tasty -package HUnit ../Network/URI.hs uri001.hs
 --  5. ./uri001
 --
---  Previous build instructions:
---  Using GHC, I compile with this command line:
---  ghc --make -fglasgow-exts
---      -i..\;C:\Dev\Haskell\Lib\HUnit;C:\Dev\Haskell\Lib\Parsec
---      -o URITest.exe URITest -main-is URITest.main
---  The -i line may need changing for alternative installations.
---
 --------------------------------------------------------------------------------
 
 module Main where
@@ -54,19 +46,15 @@ import Network.URI
 
 import Test.HUnit
 
+import Data.Bits ((.&.), (.|.))
+import Data.Char (ord, chr)
 import Data.Maybe (fromJust)
 import Data.List (intercalate)
 import System.IO (openFile, IOMode(WriteMode), hClose)
-import qualified Test.Framework as TF
-import qualified Test.Framework.Providers.HUnit as TF
-import qualified Test.Framework.Providers.QuickCheck2 as TF
-
--- Test supplied string for valid URI reference syntax
---   isValidURIRef :: String -> Bool
--- Test supplied string for valid absolute URI reference syntax
---   isAbsoluteURIRef :: String -> Bool
--- Test supplied string for valid absolute URI syntax
---   isAbsoluteURI :: String -> Bool
+import qualified Test.Tasty as TF
+import qualified Test.Tasty.HUnit as TF
+import qualified Test.Tasty.QuickCheck as TF
+import Test.QuickCheck ((==>), Property)
 
 data URIType = AbsId    -- URI form (absolute, no fragment)
              | AbsRf    -- Absolute URI reference
@@ -99,7 +87,7 @@ testURIRef t u = sequence_
   , testEq ("test_isAbsoluteURI:"++u)  (isAbsIdT t) (isAbsoluteURI  u)
   ]
 
-testURIRefComponents :: String -> (Maybe URI) -> String -> Assertion
+testURIRefComponents :: String -> Maybe URI -> String -> Assertion
 testURIRefComponents _lab uv us =
     testEq ("testURIRefComponents:"++us) uv (parseURIReference us)
 
@@ -238,19 +226,12 @@ testURIRef120 = testURIRef AbsId "http://192.168.0.1test.example.com/"
 testURIRef121 = testURIRef AbsId "http://[v9.123.abc;456.def]/"
 testURIRef122 = testEq "v.future authority"
                        (Just (URIAuth "" "[v9.123.abc;456.def]" ":42"))
-                       ((maybe Nothing uriAuthority) . parseURI $ "http://[v9.123.abc;456.def]:42/")
--- URI with non-ASCII characters, fail with Network.HTTP escaping code (see below)
--- Currently not supported by Network.URI, but captured here for possible future reference
--- when IRI support may be added.
+                       (maybe Nothing uriAuthority . parseURI $ "http://[v9.123.abc;456.def]:42/")
+-- URIs with non-ASCII characters (IRIs), are not supported by Network.URI, but
+-- captured here for possible future reference when IRI support may be added.
 testURIRef123 = testURIRef AbsId "http://example.com/test123/䡥汬漬⁗潲汤/index.html"
 testURIRef124 = testURIRef AbsId "http://example.com/test124/Москва/index.html"
 
--- From report by Alexander Ivanov:
--- should return " 䡥汬漬⁗潲汤", but returns "Hello, World" instead
--- print $ urlDecode $ urlEncode " 䡥汬漬⁗潲汤"
--- should return "Москва"
--- print $ urlDecode $ urlEncode "Москва"
-
 testURIRefSuite = TF.testGroup "Test URIrefs" testURIRefList
 testURIRefList =
   [ TF.testCase "testURIRef001" testURIRef001
@@ -393,15 +374,7 @@ testComponent01 = testURIRefComponents "testComponent01"
             } )
         "http://user:pass@example.org:99/aaa/bbb?qqq#fff"
 testComponent02 = testURIRefComponents "testComponent02"
-        ( const Nothing
-        ( Just $ URI
-            { uriScheme    = "http:"
-            , uriAuthority = Just (URIAuth "user:pass@" "example.org" ":99")
-            , uriPath      = "aaa/bbb"
-            , uriQuery     = ""
-            , uriFragment  = ""
-            } )
-        )
+        Nothing
         "http://user:pass@example.org:99aaa/bbb"
 testComponent03 = testURIRefComponents "testComponent03"
         ( Just $ URI
@@ -412,7 +385,7 @@ testComponent03 = testURIRefComponents "testComponent03"
             , uriFragment  = ""
             } )
         "http://user:pass@example.org:99?aaa/bbb"
-testComponent04 = testURIRefComponents "testComponent03"
+testComponent04 = testURIRefComponents "testComponent04"
         ( Just $ URI
             { uriScheme    = "http:"
             , uriAuthority = Just (URIAuth "user:pass@" "example.org" ":99")
@@ -421,8 +394,35 @@ testComponent04 = testURIRefComponents "testComponent03"
             , uriFragment  = "#aaa/bbb"
             } )
         "http://user:pass@example.org:99#aaa/bbb"
+testComponent05 = testURIRefComponents "testComponent05"
+        ( Just $ URI
+            { uriScheme    = "news:"
+            , uriAuthority = Nothing
+            , uriPath      = "comp.infosystems.www.servers.unix"
+            , uriQuery     = ""
+            , uriFragment  = ""
+            } )
+        "news:comp.infosystems.www.servers.unix"
+testComponent06 = testURIRefComponents "testComponent06"
+        ( Just $ URI
+            { uriScheme    = "mailto:"
+            , uriAuthority = Nothing
+            , uriPath      = "John.Doe@example.com"
+            , uriQuery     = ""
+            , uriFragment  = ""
+            } )
+        "mailto:John.Doe@example.com"
+testComponent07 = testURIRefComponents "testComponent07"
+        ( Just $ URI
+            { uriScheme    = "tel:"
+            , uriAuthority = Nothing
+            , uriPath      = "+1-816-555-1212"
+            , uriQuery     = ""
+            , uriFragment  = ""
+            } )
+        "tel:+1-816-555-1212"
 -- These test cases contributed by Robert Buck (mathworks.com)
-testComponent11 = testURIRefComponents "testComponent03"
+testComponent11 = testURIRefComponents "testComponent11"
         ( Just $ URI
             { uriScheme    = "about:"
             , uriAuthority = Nothing
@@ -431,7 +431,7 @@ testComponent11 = testURIRefComponents "testComponent03"
             , uriFragment  = ""
             } )
         "about:"
-testComponent12 = testURIRefComponents "testComponent03"
+testComponent12 = testURIRefComponents "testComponent12"
         ( Just $ URI
             { uriScheme    = "file:"
             , uriAuthority = Just (URIAuth "" "windowsauth" "")
@@ -441,11 +441,14 @@ testComponent12 = testURIRefComponents "testComponent03"
             } )
         "file://windowsauth/d$"
 
-testComponentSuite = TF.testGroup "Test URIrefs" $
+testComponentSuite = TF.testGroup "Test URIrefs"
   [ TF.testCase "testComponent01" testComponent01
   , TF.testCase "testComponent02" testComponent02
   , TF.testCase "testComponent03" testComponent03
   , TF.testCase "testComponent04" testComponent04
+  , TF.testCase "testComponent05" testComponent05
+  , TF.testCase "testComponent06" testComponent06
+  , TF.testCase "testComponent07" testComponent07
   , TF.testCase "testComponent11" testComponent11
   , TF.testCase "testComponent12" testComponent12
   ]
@@ -485,8 +488,8 @@ testRelJoin label base urel uabs =
 testRelative :: String -> String -> String -> String -> Assertion
 testRelative label base uabs urel = sequence_
     [
-    (testRelSplit (label++"(rel)") base uabs urel),
-    (testRelJoin  (label++"(abs)") base urel uabs)
+    testRelSplit (label++"(rel)") base uabs urel,
+    testRelJoin  (label++"(abs)") base urel uabs
     ]
 
 testRelative01 = testRelative "testRelative01"
@@ -624,7 +627,7 @@ testRelative55 = testRelative "testRelative55"
                     "http://example/x/y%2Fz" "http://example/x%2Fabc" "/x%2Fabc"
 -- Apparently, TimBL prefers the following way to 41, 42 above
 -- cf. http://lists.w3.org/Archives/Public/uri/2003Feb/0028.html
--- He also notes that there may be different relative fuctions
+-- He also notes that there may be different relative functions
 -- that satisfy the basic equivalence axiom:
 -- cf. http://lists.w3.org/Archives/Public/uri/2003Jan/0008.html
 testRelative56 = testRelative "testRelative56"
@@ -671,9 +674,9 @@ testRelative77 = testRelative "testRelative77"
                     "foo:bar" "http://example/a/b#c/../d"  "http://example/a/b#c/../d"
 {- These (78-81) are some awkward test cases thrown up by a question on the URI list:
      http://lists.w3.org/Archives/Public/uri/2005Jul/0013
-   Mote that RFC 3986 discards path segents after the final '/' only when merging two
-   paths - otherwise the final segment in the base URI is mnaintained.  This leads to
-   difficulty in constructinmg a reversible relativeTo/relativeFrom pair of functions.
+   Note that RFC 3986 discards path segments after the final '/' only when merging two
+   paths - otherwise the final segment in the base URI is maintained.  This leads to
+   difficulty in constructing a reversible relativeTo/relativeFrom pair of functions.
 -}
 testRelative78 = testRelative "testRelative78"
                     "http://www.example.com/data/limit/.." "http://www.example.com/data/limit/test.xml"
@@ -1071,7 +1074,7 @@ ts04str = "http://user:...@example.org:99/aaa/bbb?ccc#ddd/eee"
 
 testShowURI01 = testEq "testShowURI01" ""      (show nullURI)
 testShowURI02 = testEq "testShowURI02" ts02str (show ts02URI)
-testShowURI03 = testEq "testShowURI03" ts03str ((uriToString id ts02URI) "")
+testShowURI03 = testEq "testShowURI03" ts03str (uriToString id ts02URI "")
 testShowURI04 = testEq "testShowURI04" ts04str (show ts04URI)
 
 testShowURI = TF.testGroup "testShowURI"
@@ -1108,12 +1111,40 @@ testEscapeURIString06 = testEq "testEscapeURIString06"
     "hello%C3%B8%C2%A9%E6%97%A5%E6%9C%AC"
     (escapeURIString isUnescapedInURIComponent "helloø©日本")
 
-propEscapeUnEscapeLoop :: String -> Bool
-propEscapeUnEscapeLoop s = s == (unEscapeString $! escaped)
+-- From report by Alexander Ivanov:
+-- should return " 䡥汬漬⁗潲汤", but returns "Hello, World" instead
+-- print $ urlDecode $ urlEncode " 䡥汬漬⁗潲汤"
+assertUnescapeEscapeInverse lab x = testEq lab x (unEscapeString $ escapeURIString isUnescapedInURIComponent x)
+testUnescapeEscape01 = assertUnescapeEscapeInverse "testUnescapeEscape01" " 䡥汬漬⁗潲汤"
+-- should return "Москва"
+-- print $ urlDecode $ urlEncode "Москва"
+testUnescapeEscape02 = assertUnescapeEscapeInverse "testUnescapeEscape02" "Москва"
+
+validUnicodePoint :: Char -> Bool
+validUnicodePoint c =
+  case ord c of
+    a | a >= 0xFDD0 && a <= 0xFDEF -> False
+    a | a .&. 0xFFFE == 0xFFFE -> False
+    _ -> True
+
+propEscapeUnEscapeLoop :: String -> Property
+propEscapeUnEscapeLoop s =
+  all validUnicodePoint s ==>
+  s == (unEscapeString $! escaped)
         where
         escaped = escapeURIString (const False) s
         {-# NOINLINE escaped #-}
 
+-- Test some Unicode chars high in the Basic Multilingual Plane.
+propEscapeUnEscapeLoopHiChars :: Char -> Property
+propEscapeUnEscapeLoopHiChars c' =
+  let c = chr $ ord c' .|. 0xff00 in
+  validUnicodePoint c ==>
+  [c] == (unEscapeString $! escaped c)
+        where
+        escaped c = escapeURIString (const False) [c]
+        {-# NOINLINE escaped #-}
+
 testEscapeURIString = TF.testGroup "testEscapeURIString"
   [ TF.testCase "testEscapeURIString01" testEscapeURIString01
   , TF.testCase "testEscapeURIString02" testEscapeURIString02
@@ -1121,7 +1152,10 @@ testEscapeURIString = TF.testGroup "testEscapeURIString"
   , TF.testCase "testEscapeURIString04" testEscapeURIString04
   , TF.testCase "testEscapeURIString05" testEscapeURIString05
   , TF.testCase "testEscapeURIString06" testEscapeURIString06
+  , TF.testCase "testUnescapeEscape01" testUnescapeEscape01
+  , TF.testCase "testUnescapeEscape02" testUnescapeEscape02
   , TF.testProperty "propEscapeUnEscapeLoop" propEscapeUnEscapeLoop
+  , TF.testProperty "propEscapeUnEscapeLoopHiChars" propEscapeUnEscapeLoopHiChars
   ]
 
 -- URI string normalization tests
@@ -1176,17 +1210,17 @@ trbase = fromJust $ parseURIReference "http://bar.org/"
 testRelativeTo01 = testEq "testRelativeTo01"
     "http://bar.org/foo"
     (show $
-      (fromJust $ parseURIReference "foo") `relativeTo` trbase)
+      fromJust (parseURIReference "foo") `relativeTo` trbase)
 
 testRelativeTo02 = testEq "testRelativeTo02"
     "http:foo"
     (show $
-      (fromJust $ parseURIReference "http:foo") `relativeTo` trbase)
+      fromJust (parseURIReference "http:foo") `relativeTo` trbase)
 
 testRelativeTo03 = testEq "testRelativeTo03"
     "http://bar.org/foo"
     (show $
-      (fromJust $ parseURIReference "http:foo") `nonStrictRelativeTo` trbase)
+      fromJust (parseURIReference "http:foo") `nonStrictRelativeTo` trbase)
 
 testRelativeTo = TF.testGroup "testRelativeTo"
   [ TF.testCase "testRelativeTo01" testRelativeTo01
@@ -1330,7 +1364,7 @@ testRectify = TF.testGroup "testRectify"
   ]
 
 -- Full test suite
-allTests =
+allTests = TF.testGroup "all"
   [ testURIRefSuite
   , testComponentSuite
   , testRelativeSuite
@@ -1410,94 +1444,3 @@ cu02 = ou02 `relativeTo` bu02
 --------------------------------------------------------------------------------
 -- $Source: /srv/cvs/cvs.haskell.org/fptools/libraries/network/tests/URITest.hs,v $
 -- $Author: gklyne $
--- $Revision: 1.8 $
--- $Log: URITest.hs,v $
--- Revision 1.81 2012/08/01           aaronfriel
--- Added additional test case for the "xip.io" service style URLs and absolute URLs prefixed with ipv4 addresses.
---
--- Revision 1.8  2005/07/19 22:01:27  gklyne
--- Added some additional test cases raised by discussion on URI@w3.org mailing list about 2005-07-19.  The test p[roposed by this discussion exposed a subtle bug in relativeFrom not being an exact inverse of relativeTo.
---
--- Revision 1.7  2005/06/06 16:31:44  gklyne
--- Added two new test cases.
---
--- Revision 1.6  2005/05/31 17:18:36  gklyne
--- Added some additional test cases triggered by URI-list discussions.
---
--- Revision 1.5  2005/04/07 11:09:37  gklyne
--- Added test cases for alternate parsing functions (including deprecated 'parseabsoluteURI')
---
--- Revision 1.4  2005/04/05 12:47:32  gklyne
--- Added test case.
--- Changed module name, now requires GHC -main-is to compile.
--- All tests run OK with GHC 6.4 on MS-Windows.
---
--- Revision 1.3  2004/11/05 17:29:09  gklyne
--- Changed password-obscuring logic to reflect late change in revised URI
--- specification (password "anonymous" is no longer a special case).
--- Updated URI test module to use function 'escapeURIString'.
--- (Should unEscapeString be similarly updated?)
---
--- Revision 1.2  2004/10/27 13:06:55  gklyne
--- Updated URI module function names per:
--- http://www.haskell.org//pipermail/cvs-libraries/2004-October/002916.html
--- Added test cases to give better covereage of module functions.
---
--- Revision 1.1  2004/10/14 16:11:30  gklyne
--- Add URI unit test to cvs.haskell.org repository
---
--- Revision 1.17  2004/10/14 11:51:09  graham
--- Confirm that URITest runs with GHC.
--- Fix up some comments and other minor details.
---
--- Revision 1.16  2004/10/14 11:45:30  graham
--- Use moduke name main for GHC 6.2
---
--- Revision 1.15  2004/08/11 11:07:39  graham
--- Add new test case.
---
--- Revision 1.14  2004/06/30 11:35:27  graham
--- Update URI code to use hierarchical libraries for Parsec and Network.
---
--- Revision 1.13  2004/06/22 16:19:16  graham
--- New URI test case added.
---
--- Revision 1.12  2004/04/21 15:13:29  graham
--- Add test case
---
--- Revision 1.11  2004/04/21 14:54:05  graham
--- Fix up some tests
---
--- Revision 1.10  2004/04/20 14:54:13  graham
--- Fix up test cases related to port number in authority,
--- and add some more URI decomposition tests.
---
--- Revision 1.9  2004/04/07 15:06:17  graham
--- Add extra test case
--- Revise syntax in line with changes to RFC2396bis
---
--- Revision 1.8  2004/03/17 14:34:58  graham
--- Add Network.HTTP files to CVS
---
--- Revision 1.7  2004/03/16 14:19:38  graham
--- Change licence to BSD style;  add nullURI definition; new test cases.
---
--- Revision 1.6  2004/02/20 12:12:00  graham
--- Add URI normalization functions
---
--- Revision 1.5  2004/02/19 23:19:35  graham
--- Network.URI module passes all test cases
---
--- Revision 1.4  2004/02/17 20:06:02  graham
--- Revised URI parser to reflect latest RFC2396bis (-04)
---
--- Revision 1.3  2004/02/11 14:32:14  graham
--- Added work-in-progress notes.
---
--- Revision 1.2  2004/02/02 14:00:39  graham
--- Fix optional host name in URI.  Add test cases.
---
--- Revision 1.1  2004/01/27 21:13:45  graham
--- New URI module and test suite added,
--- implementing the GHC Network.URI interface.
---
